/// \file TiffFile.cpp
/// implements CTiffFile class

/***************************************************************************
 *  Copyright (C) 2007 Pascal Corpet
 *  http://corpet.net/Pascal
 *
 *  This file is part of TiffRectoVerso.
 *
 *  TiffRectoVerso is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  TiffRectoVerso is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with TiffRectoVerso; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
***************************************************************************/

#include "TiffFile.h"
#include "ImageFileDirectory.h"

#include <crtdbg.h>

#define ASSERT(a) _ASSERT(a)
#if _DEBUG
	#define VERIFY(a) _ASSERT(a)
#else
	#define VERIFY(a) (a)
#endif

CTiffFile::CTiffFile(_TCHAR* i_pccFileName) :
	m_szNbIFDs(0),
	m_allIfds(NULL)
{
	if (_tfopen_s(&m_file, i_pccFileName, _T("rb+")) != 0)
	{
		m_file = NULL;
		return;
	}
	uint16 usBuf;
	if (fread(&usBuf, sizeof(usBuf), 1, m_file) != 1 ||
		(usBuf != 0x4949 && usBuf != 0x4D4D))
	{
		// this is not a tif file
		fclose(m_file);
		return;
	}
	m_bLittleEndian = usBuf == 0x4949;

	try
	{
		VERIFY(ReadUInt16() == 42);
	}
	catch (...)
	{
		// file is to short
		m_file = NULL;
		return;
	}
}

uint16 CTiffFile::ReadUInt16()
{
	ASSERT(!Closed());
	uint16 a;
	if (fread(&a, sizeof(a), 1, m_file) != 1)
	{
		throw "Impossible to read an UInt16 from file";
	}
	if (!m_bLittleEndian)
	{
		uint8* b = reinterpret_cast<uint8*>(&a);
		uint8 c = b[0];
		b[0] = b[1];
		b[1] = c;
	}
	return a;
}

uint32 CTiffFile::ReadAddress()
{
	uint32 address;
	if (fread(&address, sizeof(address), 1, m_file) != 1)
	{
		throw "Impossible to read an address from file";
	}
	if (!m_bLittleEndian)
	{
		uint8* b = reinterpret_cast<uint8*>(&address);
		uint8 c = b[0];
		b[0] = b[3];
		b[3] = c;
		c = b[1];
		b[1] = b[2];
		b[2] = c;
	}
	return address;
}

bool CTiffFile::WriteAddress(uint32 i_address)
{
	uint32 address = i_address;
	if (!m_bLittleEndian)
	{
		uint8* b = reinterpret_cast<uint8*>(&address);
		uint8 c = b[0];
		b[0] = b[3];
		b[3] = c;
		c = b[1];
		b[1] = b[2];
		b[2] = c;
	}
	if (fwrite(&address, sizeof(address), 1, m_file) != 1)
	{
		throw "Impossible to write an address to file";
	}
	return true;
}

int CTiffFile::SeekToAddress(uint32 i_address)
{
	ASSERT(!Closed());
	uint32 address = i_address;
	return fseek(m_file, address, SEEK_SET);
}

CTiffFile::~CTiffFile()
{
	if (m_allIfds != NULL)
	{
		for (size_t i = 0; i < m_szNbIFDs; i++)
		{
			delete m_allIfds[i];
		}
		delete [] m_allIfds;
		m_allIfds = NULL;
	}
	if (m_file != NULL)
	{
		fclose(m_file);
		m_file = NULL;
	}
}

bool CTiffFile::Closed() const
{
	return m_file == NULL;
}

size_t CTiffFile::ReadAllDirectories()
{
	if (m_allIfds != NULL)
	{
		// we already read the directories
		return m_szNbIFDs;
	}
	CImageFileDirectory* firstIfd = new CImageFileDirectory(*this);
	ASSERT(!firstIfd->End());

	size_t szNbIfd = 1;
	CImageFileDirectory* endIfd;
	for (endIfd = firstIfd; !endIfd->End(); endIfd = new CImageFileDirectory(*this, endIfd))
	{
		szNbIfd++;
	}
	szNbIfd--;

	m_allIfds = new CImageFileDirectory*[szNbIfd];
	m_allIfds[szNbIfd-1] = endIfd->Previous();
	delete (endIfd);

	for (size_t i = szNbIfd-1; i > 0; i--)
	{
		m_allIfds[i-1] = m_allIfds[i]->Previous();
	}

	m_szNbIFDs = szNbIfd;
	return m_szNbIFDs;
}

int CTiffFile::MixRectoAndVerso()
{
	// we have all ifds in one list, now we assume that this list is in a wrong order :
	// 1 3 5 7 9 ... (2n+1) [(2n+2)] (2n) ... 8 6 4 2
	// the (2n+2) is there only if the number of allIfds is even

	for (size_t k = 0; 2*k < m_szNbIFDs-2; k++)
	{
		m_allIfds[k]->InsertAfterMe(*this,m_allIfds[m_szNbIFDs-1-k]);
	}

	return 0;
}

int CTiffFile::SplitRectoAndVerso()
{
	// we have all ifds in one list, now we want them in a new order :
	// 1 3 5 7 9 ... (2n+1) [(2n+2)] (2n) ... 8 6 4 2
	// the (2n+2) is there only if the number of allIfds is even

	for (size_t k = 1; k < m_szNbIFDs-1; k+=2)
	{
		m_allIfds[m_szNbIFDs-1]->InsertAfterMe(*this,m_allIfds[k]);
	}

	return 0;
}
